import math

from ..base import BaseScheduler, register_scheduler


@register_scheduler("polynomial")
class PolynomialScheduler(BaseScheduler):
    def _sub_init(self, power=1, closed_form=True):
        self.power = power
        self.closed_form = closed_form

    def _get_value(self, step):
        decay = (
            (1 - (step / self.period))
            / (1 if self.closed_form else (1 - (step - 1) / self.period))
        ) ** self.power
        return (self.value - self.min_value) * decay + self.min_value



import math

from ..base import BaseScheduler, register_scheduler


@register_scheduler("cosine")
class CosineScheduler(BaseScheduler):
    def _get_value(self, step):
        if step <= 0:
            return self.value
        if step >= self.period:
            return self.min_value
        return self.min_value + 0.5 * (self.value - self.min_value) * (
            1 + math.cos(math.pi * step / self.period)
        )


if __name__ == "__main__":
    scheduler = CosineScheduler(0, 100, warmup=10, value=0.1, min_value=0.01)
    for step in range(100):
        print(f"step {step}: {scheduler(step)}")



from typing import Any
from pathlib import Path
from copy import deepcopy

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
import toml

from .utils import get_scheduler


class AnySchedule(lr_scheduler.LRScheduler):
    def __init__(
        self,
        optimizer: optim.Optimizer,
        last_epoch: int = -1,
        verbose="deprecated",
        config: str | dict[str, Any] = {},
    ):
        if isinstance(config, (str, Path)):
            config = toml.load(config)
        self.config = deepcopy(config)
        self.schedulers = {key: get_scheduler(val) for key, val in self.config.items()}
        self.base_param_groups = {
            key: [group[key] for group in optimizer.param_groups]
            for key in self.schedulers
        }
        super(AnySchedule, self).__init__(optimizer, last_epoch)

    def state_dict(self):
        """Return the state of the scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in {"optimizer", "schedulers"}
        }

    def get_factors(self):
        return {
            key: scheduler(self.last_epoch)
            for key, scheduler in self.schedulers.items()
        }

    def get_lr(self):
        if "lr" in self.schedulers:
            factor = self.schedulers["lr"](self.last_epoch)
            return [val * factor for val in self.base_param_groups["lr"]]
        else:
            return self.base_param_groups["lr"]

    def step(self):
        super(AnySchedule, self).step()
        for key, scheduler in self.schedulers.items():
            if key == "lr":
                continue
            factor = scheduler(self.last_epoch)
            for group, val in zip(self.optimizer.param_groups, self.base_param_groups[key]):
                if key not in group:
                    continue
                if isinstance(group[key], torch.Tensor):
                    group[key].fill(factor * val)
                else:
                    group[key] = factor * val



class BaseScheduler:
    """
    HyperParameter scheduler base class
    start: int, start step, include
    end: int, end step, exclude
    warmup: int, warmup steps
    value: float, initial value

    _get_value: return the value of hyperparameter at "internal" step
    __call__: return the value of hyperparameter at "global" step
    """

    def __init__(
        self,
        start=None,
        end=None,
        warmup=0,
        value=1.0,
        min_value=None,
        init_value=None,
        step_size=1,
        **kwargs,
    ):
        assert value is not None, "value must be provided"
        self.value = value
        self.init_value = init_value
        self.min_value = min_value or 0
        self.start = start or 0
        self.warmup = warmup or 0
        self.end = end or float("inf")
        self.period = (self.end - self.start - self.warmup) // step_size
        self.step_size = step_size

        if self.end == float("inf"):
            self.period = float("inf")

        assert (
            self.period > 0
        ), f"Invalid settings: start={start}, end={end}, warmup={warmup}"
        self._sub_init(**kwargs)

    def _sub_init(self):
        pass

    def _get_value(self, step):
        raise NotImplementedError

    def __call__(self, step):
        step -= self.start
        if step >= self.end:
            raise ValueError(f"Step {step} is out of range [{self.start}, {self.end})")
        if step < self.warmup:
            interp = step / self.warmup
            init_val = self.init_value or 0
            return self.value * interp + init_val * (1 - interp)
        step -= self.warmup
        step //= self.step_size
        return max(self.min_value, self._get_value(step))


registered_schedulers = {}


def register_scheduler(name):
    def register(scheduler):
        if not issubclass(scheduler, BaseScheduler):
            raise ValueError(
                f"Only subclass of BaseScheduler can be registered, got {scheduler}"
            )
        registered_schedulers[name] = scheduler
        return scheduler

    return register



from typing import Any, Optional
from .base import BaseScheduler, register_scheduler, registered_schedulers
from .scheduler.constant import ConstantScheduler


@register_scheduler("composer")
class Composer(BaseScheduler):
    def _sub_init(
        self,
        schedules: dict[str, Any] = {},
    ):
        self.steps_schedule = []
        st = 0
        final_value = self.value
        for idx, schedule in enumerate(schedules):
            if isinstance(schedule, dict):
                if "start" not in schedule:
                    schedule["start"] = st
                if "value" not in schedule:
                    schedule["value"] = final_value
                schedule_cls = registered_schedulers[schedule.pop("mode")]
                schedules[idx] = schedule_cls(**schedule)
            elif not isinstance(schedule, BaseScheduler):
                raise ValueError(
                    f"Only subclass of BaseScheduler or config dict can be used, got {schedule}"
                )
            current_schedule = schedules[idx]
            if current_schedule.init_value is None:
                current_schedule.init_value = final_value
            start = current_schedule.start
            end = current_schedule.end
            st = end
            final_value = current_schedule(end - 1)
            if self.steps_schedule and start > self.steps_schedule[-1][1]:
                self.steps_schedule.append((self.steps_schedule[-1][1], start, -1))
            self.steps_schedule.append((start, end, idx))
        self.steps_schedule.append((self.steps_schedule[-1][1], float("inf"), -1))
        self.schedules = schedules + [ConstantScheduler(value=self.value)]
        self.current_schedule = self.schedules[0]
        self.current_start = self.steps_schedule[0][0]
        self.current_end = self.steps_schedule[0][1]

    def find_range(self, step: int) -> int:
        left, right = 0, len(self.steps_schedule) - 1

        while left <= right:
            mid = (left + right) // 2
            start, end, idx = self.steps_schedule[mid]

            if start <= step < end:
                return start, end, idx
            elif step >= end:
                left = mid + 1
            else:
                right = mid - 1
        return -1, -1, -1

    def _get_value(self, step):
        if step < self.current_start or step >= self.current_end:
            self.current_start, self.current_end, idx = self.find_range(step)
            self.current_schedule = self.schedules[idx]
        return self.current_schedule(step)



from ..base import BaseScheduler, register_scheduler


@register_scheduler("step")
class StepScheduler(BaseScheduler):
    def _sub_init(self, gamma=0.95):
        self.gamma = gamma

    def _get_value(self, step):
        return self.value * (self.gamma**step)
